#include <stdafx.h>


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <memory.h>


#include "wos.h"
#include "deb.h"
#include "fsx.h"


#include "zlib.h"

#include "zip.h"


#include "deb.h"



static char          * zstart;
static char          * zend;
static char          * zenddir;
static char          * zcentdir;
static unsigned int    zlength;


#define FILE_SIGNATURE 0x04034b50
#define CENTRAL_SIGNATURE 0x02014b50
#define END_SIGNATURE 0x06054b50



typedef unsigned char uint8;
typedef unsigned int  uint32;
typedef unsigned short uint16;




typedef struct file_header
{
 uint16 version_needed_to_extract;
 uint16 general_purpose_bit_flag;
 uint16 compression_method;
 uint16 last_mod_file_time;
 uint16 last_mod_file_date;
 uint32 crc32;
 uint32 compressed_size;
 uint32 uncompressed_size;
 uint32 filename_length;
 uint32 extra_field_length;
} file_header;


#define file_header_size 26
#define version_needed_to_extract_offset 0
#define general_purpose_bit_flag_offset 2
#define compression_method_offset 4
#define last_mod_file_time_offset 6
#define last_mod_file_date_offset 8
#define crc32_offset 10
#define compressed_size_offset 14
#define uncompressed_size_offset 18
#define filename_length_offset 22
#define extra_field_length_offset 24





typedef struct central_header
{
 uint16 version_made_by;
 uint16 version_needed_to_extract;
 uint16 general_purpose_bit_flag;
 uint16 compression_method;
 uint16 last_mod_file_time;
 uint16 last_mod_file_date;
 uint32 crc32;
 uint32 compressed_size;
 uint32 uncompressed_size;
 uint16 filename_length;
 uint16 extra_field_length;
 uint16 file_comment_length;
 uint16 disk_number_start;
 uint16 internal_file_attributes;
 uint32 external_file_attributes;
 uint32 relative_offset_local_header;
 char   extra[4];

} central_header;


#define central_header_size 42

#define c_version_made_by_offset 0
#define c_version_needed_to_extract_offset 2
#define c_general_purpose_bit_flag_offset 4
#define c_compression_method_offset 6
#define c_last_mod_file_time_offset 8
#define c_last_mod_file_date_offset 10
#define c_crc32_offset 12
#define c_compressed_size_offset 16
#define c_uncompressed_size_offset 20
#define c_filename_length_offset 24
#define c_extra_field_length_offset 26
#define c_file_comment_length_offset 28
#define c_disk_number_start_offset 30
#define c_internal_file_attributes_offset 32
#define c_external_file_attributes_offset 34
#define c_relative_offset_local_header_offset 38





typedef struct end_header
{
 uint16 number_this_disk;
 uint16 number_disk_with_start_central_directory;
 uint16 total_entries_central_dir_on_this_disk;
 uint16 total_entries_central_dir;
 uint32 size_central_directory;
 uint32 offset_start_central_directory;
 uint16 zipfile_comment_length;
 char   comment[4];
} end_header;


#define end_header_size 18

#define e_number_this_disk_offset 0
#define e_number_disk_with_start_central_directory_offset 2
#define e_total_entries_central_dir_on_this_disk_offset 4
#define e_total_entries_central_dir_offset 6
#define e_size_central_directory_offset 8
#define e_offset_start_central_directory_offset 12
#define e_zipfile_comment_length_offset 16


uint16 getuint16(uint8 * p)
{
 return((*p)+((*(p+1))<<8));
}

uint32 getuint32(uint8 * p)
{
 return((*p)+((*(p+1))<<8)+((*(p+2))<<16)+((*(p+3))<<24));
}


void putuint16(uint8 * p,uint16 u)
{
 *p++=u & 0xFF;
 *p=(u>>8) & 0xFF;
}


void putuint32(uint8 * p,uint32 u)
{
 *p++=u & 0xFF;
 *p++=(u>>8) & 0xFF;
 *p++=(u>>16) & 0xFF;
 *p=(u>>24) & 0xFF;

}





// is first the lhs of second?

static int fstrncmp(char * first,char * second,int n)
{
 int f;
 int s;

 while(n--)
 {
  f=toupper(*first++);
  s=toupper(*second++);
  if(f=='/') f='\\';
  if(s=='/') s='\\';
  if(!f) return(0);
  if(f!=s) return(f-s);
 }

	return(1);
 return(*first);
}



static void fnamecpy(char * d,char * s,int len)
{
 int ch;

 while(len--)
 {
  ch=*s++;
  if(ch=='/') ch='\\';
  *d++=ch;
 }

 *d=0;
}




static void process_central_header(char * p,int * lenp,char * name,char ** localp,char * found)
{
 central_header   chdr;
 uint8            hdr[central_header_size];
 int              len;


 memcpy(hdr,p,central_header_size);


 *lenp=(central_header_size);

 cprintf("process central");

 	chdr.version_made_by=getuint16(hdr+c_version_made_by_offset);
  chdr.version_needed_to_extract=getuint16(hdr+c_version_needed_to_extract_offset);
  chdr.general_purpose_bit_flag=getuint16(hdr+c_general_purpose_bit_flag_offset);
  chdr.compression_method=getuint16(hdr+c_compression_method_offset);
  chdr.last_mod_file_time=getuint16(hdr+c_last_mod_file_time_offset);
  chdr.last_mod_file_date=getuint16(hdr+c_last_mod_file_date_offset);
  chdr.crc32=getuint32(hdr+c_crc32_offset);
  chdr.compressed_size=getuint32(hdr+c_compressed_size_offset);
  chdr.uncompressed_size=getuint32(hdr+c_uncompressed_size_offset);
  chdr.filename_length=getuint16(hdr+c_filename_length_offset);
  chdr.extra_field_length=getuint16(hdr+c_extra_field_length_offset);
  chdr.file_comment_length=getuint16(hdr+c_file_comment_length_offset);
  chdr.disk_number_start=getuint16(hdr+c_disk_number_start_offset);
  chdr.internal_file_attributes=getuint16(hdr+c_internal_file_attributes_offset);
  chdr.external_file_attributes=getuint32(hdr+c_external_file_attributes_offset);
  chdr.relative_offset_local_header=getuint32(hdr+c_relative_offset_local_header_offset);


 if(name==NULL || fstrncmp(name,p+(*lenp),chdr.filename_length)==0)
 {
//	 	char temp[256];
//   fnamecpy(temp,p+(*lenp),chdr.filename_length);
//   temp[chdr.filename_length]=0;
//			printf("match %s %s",temp,name);

  *localp=zstart+chdr.relative_offset_local_header;
  if(found)
  {
   fnamecpy(found,p+(*lenp),chdr.filename_length);
   found[chdr.filename_length]=0;

  }
 }

 cprintf("reloffset %d filenamelen %d",chdr.relative_offset_local_header,chdr.filename_length);

 len =chdr.filename_length;
 len+=chdr.extra_field_length;
 len+=chdr.file_comment_length;

 *lenp+=len;
}










static void processenddir(char * p)
{
 end_header        ehdr;
 uint8             hdr[end_header_size];

 cprintf("process end dir");

 memcpy(hdr,p,end_header_size);

 ehdr.number_this_disk=getuint16(hdr+e_number_this_disk_offset);
 ehdr.number_disk_with_start_central_directory=getuint16(hdr+e_number_disk_with_start_central_directory_offset);
 ehdr.total_entries_central_dir_on_this_disk=getuint16(hdr+e_total_entries_central_dir_on_this_disk_offset);
 ehdr.total_entries_central_dir=getuint16(hdr+e_total_entries_central_dir_offset);
 ehdr.size_central_directory=getuint32(hdr+e_size_central_directory_offset);
 ehdr.offset_start_central_directory=getuint32(hdr+e_offset_start_central_directory_offset);
 ehdr.zipfile_comment_length=getuint16(hdr+e_zipfile_comment_length_offset);
  
 zcentdir=zstart+ehdr.offset_start_central_directory;
}



// name - what we're looking for
// start - pointer to where to start looking
// found name of what we found
// next - where to look next
// returns pointer to local header in zip file



char * zipseek(char * name,char * start,char * found,char ** next)
{
 char    * p;
 char    * q;
 int       sig;
 int       len;

 if(start==NULL) p=zcentdir;
 else            p=start;

 q=NULL;
 if(next) *next=NULL;

	cprintf("zip seek %s",name);

cprintf("p %x end %x",p,zend);
 
 while(p<zend)
 {
  sig=getuint32((uint8*)p);
  p+=4;

cprintf("sig %x",sig);

  if(sig==CENTRAL_SIGNATURE)
  {
   process_central_header(p,&len,name,&q,found);
   p+=len;
   if(q)
   {
    if(next) *next=p;
    break;
   }
  }
  else
  {
   break;
  }
 }

 return(q);
}




static void read_local_file_header(char * p,file_header * fhdr)
{
 uint8 hdr[file_header_size];


 memcpy(hdr,p,(file_header_size));

 fhdr->version_needed_to_extract=getuint16(hdr+version_needed_to_extract_offset);
 fhdr->general_purpose_bit_flag=getuint16(hdr+general_purpose_bit_flag_offset);
 fhdr->compression_method=getuint16(hdr+compression_method_offset);
 fhdr->last_mod_file_time=getuint16(hdr+last_mod_file_time_offset);
 fhdr->last_mod_file_date=getuint16(hdr+last_mod_file_date_offset);
 fhdr->crc32=getuint32(hdr+crc32_offset);
 fhdr->compressed_size=getuint32(hdr+compressed_size_offset);
 fhdr->uncompressed_size=getuint32(hdr+uncompressed_size_offset);
 fhdr->filename_length=getuint16(hdr+filename_length_offset);
 fhdr->extra_field_length=getuint16(hdr+extra_field_length_offset);
}





static os_error * zipwritefilesub(char * local,char * name)
{
 os_error          * err;
 file_header         fhdr;
 int                 sig;
 char              * datap;
 int                 len;
 fshandle            fh;
 FILETIME            modtime;




 sig=getuint32((uint8*)local);
 local+=4;
 if(sig==FILE_SIGNATURE)
 {
  read_local_file_header(local,&fhdr);

  datap=local+file_header_size+fhdr.filename_length+fhdr.extra_field_length;
  len=fhdr.uncompressed_size;

cprintf("lrec.compression_method %d len=%d size %d",fhdr.compression_method,len,fhdr.compressed_size);
cprintf("zipopen %s",name);
cprintf("flen %d exlen %d",fhdr.filename_length,fhdr.extra_field_length);

  err=fs_open(name,'w',&fh);
  if(!err)
  {
   if(fhdr.compression_method==0)
   {
    err=fs_write(fh,datap,len);
   }
   else
   {
    int      code;
    z_stream stream;
    char     buffer[0x8000];
    int      wlen;
    int      left;


    left=len;
    memset(&stream,0,sizeof(z_stream));

    stream.next_in=(unsigned char*)datap;
    stream.avail_in=fhdr.compressed_size;
    stream.next_out=(unsigned char*)buffer;
    stream.avail_out=sizeof(buffer);

    code=inflateInit2(&stream,-15);
    if(code!=Z_OK) err=generror(0,"Decompression setup failed code %d",code);

    if(!err)
    {
     while(1)
     {
      code=inflate(&stream, Z_SYNC_FLUSH);

      if(code==Z_OK || code==Z_STREAM_END)
      {
       wlen=stream.next_out-(unsigned char*)buffer;
       err=fs_write(fh,buffer,wlen);
       left-=wlen;
      }
      else
      {
       err=generror(0,"Decompression failed code %d",code);
      }

      if(err) break;
      if(code!=Z_OK) break;
      if(left<=0) break;
  
      stream.next_out=(unsigned char*)buffer;
      stream.avail_out=sizeof(buffer);
     }
    }
    inflateEnd(&stream);
   }
   
   if(!err)
   {
    DosDateTimeToFileTime(fhdr.last_mod_file_date,fhdr.last_mod_file_time,&modtime);
    SetFileTime((HANDLE)fh,NULL,NULL,&modtime);
   }
   err=fs_close(fh,err);
  }
 }
 return(err);
}


os_error * zipwritefile(char * localname,char * path,char * local)
{
 os_error * err;
// char     * local;
 char       name[MAX_PATH];
 char 				* p;

	err=NULL;

 cprintf("zipwritefile");

 if(!local) local=zipseek(localname,NULL,NULL,NULL);
 if(local)
 {
  strcpy(name,path);
  strcat(name,"\\");
  strcat(name,localname);

  err=fs_validatepath(name,0);
  if(!err)
		{
   p=name+strlen(name);

			cprintf("zipwritefile name %s",name);
   if(*(p-1)=='\\' || *(p-1)=='/')
			{

			}
			else
			{
				cprintf("write");

    err=zipwritefilesub(local,name);
			}
		}
 }
 return(err);
}





os_error * zipwritefolder(char * foldername,char * path)
{
 os_error * err;
 char       name[MAX_PATH];
 char     * local;
 char     * next;
 int        len;
 char     * p;


 err=NULL;

cprintf("zipwrite folder %s %s",foldername,path);

 next=NULL;
 len=strlen(path);
 strcpy(name,path);
 if(len!=0 && name[len-1]!='\\')
 {
  name[len]='\\';
  len++;
 }

 while(1)
 {
  local=zipseek(foldername,next,name+len,&next);
  if(!next) break;

cprintf("name %s",name);

  p=name+strlen(name);
  if(*(p-1)=='\\' || *(p-1)=='/')
  {
   err=fs_validatepath(name,0);
   if(err) break;
  }
  else
  {
   err=zipwritefilesub(local,name);
   if(err) break;
  }
 }

 return(err);
}





os_error * zipopendata(char * localname,char ** datap,int * len)
{
 char              * local;
 file_header         fhdr;
 int                 sig;

 local=zipseek(localname,NULL,NULL,NULL);

 sig=getuint32((uint8*)local);
 
 local+=4;
 if(sig==FILE_SIGNATURE)
 {
  read_local_file_header(local,&fhdr);

  *datap=local+file_header_size+fhdr.filename_length+fhdr.extra_field_length;
  *len=fhdr.uncompressed_size;
 }
 return(NULL);
}


os_error * zipclosedata(char ** datap)
{


 *datap=NULL;

 return(NULL);
}


os_error * zipinit(char * start,int length)
{
 os_error * err;
 char     * p;
 int        sig;

cprintf("zipinit");

 zstart=start;
 zlength=length;
 zenddir=NULL;

 err=NULL;
 
 if(!err)
 {
  zend=zstart+zlength;

  p=zend-4;
  while(p>=zstart)
  {
   sig=getuint32((uint8*)p);
   if(sig==END_SIGNATURE)
   { 
    zenddir=p;
    break;
   }
   p--;
  }

  processenddir(zenddir+4);

 }
 return(err);
}





os_error * ziploadfile(char * path)
{
 os_error * err;
 fstat      fs;
 char     * data;

 err=NULL;

 fs_stat(path,&fs);
 if(fs.object==1)
 {
  data=(char*)malloc(fs.length);
  if(data)
  {
   err=fs_loadblock(path,data);
   if(!err) err=zipinit(data,fs.length);
  }
  else
  {
   err=generror(0,"Failed to allocate enough memory");
  }
 }

 return(err);
}

