// TextPal v1.01
//
// A demonstration OvationPro applet.
// Originally accompanied a series of Acorn User articles.
// Produces a palette that allows the leading, vertical shift and tracking
// of the text to be changed.
//
// Because of the origin of this applet as a tutorial, there are lots of
// comments in the code. Hope they are helpful!
//
// Thanks to David Pilling for his help in fathoming some of the more obscure
// aspects of the script language.
//
// John Pettigrew
// john@xl-cambridge.demon.co.uk


// First, declare all global variables

// view textpal is attached to
int textpal_view;
// textpal_handle will store the window handle of the palette window
int textpal_handle;
// textpal_centre controls whether or not the window will be centred when first opened (1 = centred)
int textpal_centre = 1;
// textpal_open records whether or not the palette window is open (1 = open)
int textpal_open = 0;
// textpal_views records the number of views open
int textpal_views = 0;
// textpal_shaded stores the shadedness of the palette (1=shaded)
int textpal_shaded = 0;
// Declare the variables to store the values of the leading & vertical shift
// First, integers to hold the numbers of the leading and shift
int textpal_lval;
int textpal_vval = 0;
// A special variable to record whether the leading is relative or absolute
// This will be prefixed to the string when setting the leading
string textpal_lrel = "";
// A string variable to hold the leading units
string textpal_lunit = "pt";
// A string variable to hold the V shift units
string textpal_vunit = "%";

int textpal_x0;
int textpal_y1;


//Function to set the value of the leading from the passed string.
//This value is set in the global variables so no need to return anything.
void textpal_leading(string leading)
{
  int len, from;
  string s;
  len = slen(leading);
  // Store in s the last character of the leading string
  s = mids(leading, len-1, 1);
  // Check whether the leading string ends with % or pt before setting values
  if(s == "%")  // If the leading's in %, set the variables appropriately
  {
    textpal_lval = stoi(mids(leading, 0, len-1));  // Extract the value
    textpal_lunit = "%";  // Set the units
    textpal_lrel = "";  // Percentage leading is an absolute measure
  }
  else
  {
    // The leading is given in points
    // However, it can be in either the "+6pt" or the "16pt" form
    // We need to work out which form is being used
    // Now, store in s the _first_ character of the leading string
    s = mids(leading, 0, 1);
    if(s == "+")  //  If it's plus, we can set the variables this way
    {
      textpal_lval = stoi(mids(leading, 1, len-3));  // Extract the value
      textpal_lunit = "pt";  // Set the units
      textpal_lrel = "+";  // Relative leading is on
    }
    else  // i.e. the first character is either a minus or a digit
    {
      if(s == "-") // Minus signs get doubled here by some versions of OvPro
      {            // so we need to take account of that
        if(mids(leading, 1, 1) == "-")  // If there are 2 minuses
        {
          from = 2; // Read the value starting with the third character
          len -= 4; // and finishing two characters before the end
        }
        else  // For versions that have this bug sorted out (2.56 or later)
        {
          from = 1; // Read the value starting with the second character
          len -= 3; // and finishing two characters before the end
        }
        textpal_lval = stoi(mids(leading, from, len));  // Extract the value
        textpal_lval = -textpal_lval;  // Make the value negative
        textpal_lunit = "pt";  // Set the units
        textpal_lrel = "-"; // Relative leading is on
      }
      else  // it's in the "16pt" form, having excluded everything else
      {
        textpal_lval = stoi(mids(leading, 0, len - 2));  // Extract the value
        textpal_lunit = "pt";  // Set the units
        textpal_lrel = ""; // Relative leading is off
      }
    }
  }
}

// Function that actually updates the palette
// It reads the current text settings and sets the palette accordingly

void textpal_update(int file)
{
 string leading;

 if(textpal_handle)
 {
  setcurrentfile(file); // Change the current file to the new one
  // Read the vshift value and correct to give 100 for 100%
  textpal_vval = getbaseshift()/655;
  getleading(leading);      // Get the current leading (as a string)
  textpal_leading(leading); // Set the global variables to the leading values
  // Set the icons to contain the values of the leading and vshift
  writeicon(textpal_handle, 21, leading);
  writeicon(textpal_handle, 5, itos(textpal_vval)+"%");
  // Then get the icons redrawn
  //refresh_area(0, textpal_handle, 214, -26, 326, -144);
  // Make sure the view is updated
  //viewrepaint(textpal_view,file);
 }
}



void textpal_open3(int h,int x0,int y0,int x1,int y1,int scx,int scy,int bh)
{
 if(!textpal_centre)
 {
  x1+=(textpal_x0-x0);
  y0+=(textpal_y1-y1);
  x0=textpal_x0;
  y1=textpal_y1;
 }
 else
 {
  textpal_x0=x0;
  textpal_y1=y1;
 }
 open_window(h,x0,y0,x1,y1,scx,scy,bh);
}


void textpal_open5(int h,int x0,int y0,int x1,int y1,int scx,int scy,int bh)
{
 textpal_x0=x0;
 textpal_y1=y1;
}




void textpal_openwindow(int view)
{
 if(view)
 {
  //Create the palette window
  textpal_handle=create_windowparent("textpal",viewhandle(view));
  setwindowflags(textpal_handle,8,8);
  //Add handlers for the palette window

  addwindowhandler(0,textpal_handle,"textpal_click");   // An icon clicked
  addwindowhandler(1,textpal_handle,"textpal_keypress");// A key pressed
  addwindowhandler(2,textpal_handle,"textpal_closepal");// Close icon click
  addwindowhandler(3,textpal_handle,"textpal_open3");
  addwindowhandler(5,textpal_handle,"textpal_open5");

  textpal_view=view;

  display_window(textpal_handle, 0, textpal_centre);
  textpal_centre=0;  // After this, the window won't be centred when opened

  textpal_update(filefromview(textpal_view));  //Update the palette to match the view
 }
}



void textpal_closewindow(void)
{
 close_window(textpal_handle);
 delete_window(textpal_handle);
 textpal_view=0;
 textpal_handle=0;

}




// Function called when the text at the cursor changes
// It calls the update function if the palette's open
void textpal_textupdate(int user, int file)
{
 if(textpal_open) textpal_update(file);
}


// Open the text palette, centring it first time only
void textpal_openpalette(void)
{
 textpal_openwindow(activeview());
 textpal_open=1;    // So that we know the palette's open
}

// Close text palette
void textpal_closepalette(void)
{
 textpal_closewindow();
 textpal_open=0;
}

// Run when the 'Text palette' entry in the Applets menu is selected
// Opens or closes the text palette window

int textpal_entry(int entry, int subcode)
{
 textpal_open=!textpal_open;
 if(!textpal_open) textpal_closepalette();
 else              textpal_openpalette();
 return(0);
}

// Tick or untick the 'Text palette' menu entry

int textpal_tick(int entry, string &text)
{
 return(textpal_open);  // 1 when open, 0 when closed
}

// Handles the close icon of the window

void textpal_closepal(int handle)
{
 textpal_closepalette();
}


// Function called when the caret enters or leaves a view
void textpal_caret(int user, int view, int gain)
{
 int loop;

 if(textpal_handle)
 {
//  for(loop=0;loop<12;loop++) set_icon_state(textpal_handle,loop,(1 << 22)*!gain, (1 << 22));
  textpal_shaded=!gain;
 }
}


// Function called when the palette window is clicked
void textpal_click(int handle, int icon, int buttons, int mousex, int mousey)
{
  int amount = bbc_inkey(-1)?1:10; // Set the amount depending on whether Shift was pressed
  // First, if we've entered the palette, we want the icons unshaded
  if(handle == textpal_handle && textpal_shaded) textpal_caret(0, 0, 1);
  // Now, deal with the click, depending on the mouse button and icon clicked
  if(buttons == 1) amount = -amount; // Reverse action if adjust was pressed
  if(buttons != 2) // i.e. if the button was not Menu
  {
    switch(icon) // Deal with the event differently depending on the parameter 'icon'
    {
      // First, leading: decrease or increase leading by amount
      case 22:
        // If we're in absolute leading, ensure it doesn't go below zero
        if(textpal_lrel == "") setleading(itos((((textpal_lval-amount) < 0) ? 0 : textpal_lval-amount))+textpal_lunit);
        // If we're in relative leading, set it and ensure the sign's correct
        else
        {
          if(textpal_lval-amount < 0) textpal_lrel = "-";
          else textpal_lrel = "+";
          setleading(textpal_lrel+itos(textpal_lval-amount)+textpal_lunit);
        }
        break;
      case 23:
        // If we're in absolute leading, ensure it doesn't go below zero
        if(textpal_lrel == "") setleading(itos((((textpal_lval+amount) < 0) ? 0 : textpal_lval+amount))+textpal_lunit);
        // If we're in relative leading, set it and ensure the sign's correct
        else
        {
          if(textpal_lval+amount < 0) textpal_lrel = "-";
          else textpal_lrel = "+";
          setleading(textpal_lrel+itos(textpal_lval+amount)+textpal_lunit);
        }
        break;
      // Then, the arrows for vshift: decrease or increase vshift by amount
      case 6:textpal_vval = textpal_vval - amount; // Set the global variable
        setbaseshift((textpal_vval)*65536/100);
        break;
      case 7:textpal_vval = textpal_vval + amount; // Set the global variable
        setbaseshift((textpal_vval)*65536/100);
        break;
      // Then the tracking, depending on the icon, mouse button and Shift key
      // The typed characters expand to keyboard shortcuts
      case 10:
        if (buttons == 4) type(bbc_inkey(-1)?"{C_;}":"{CS_;}");
        else type(amount==1?"{C_'}":"{CS_'}");
        break;
      case 11:
        if (buttons == 4) type(bbc_inkey(-1)?"{C_'}":"{CS_'}");
        else type(amount==1?"{C_;}":"{CS_;}");
        break;
    }
  }
}



void textpal_activeview(int user,int view,int gain)
{

 if(gain && textpal_open && view && view!=textpal_view)
 {
  textpal_openwindow(view);
 }

 if(!gain && textpal_open && view && view==textpal_view && textpal_handle)
 {
  textpal_closewindow();
 }

}






//Function to handle keypresses in the palette
int textpal_keypress(int handle, int icon, int key)
{
  string leading, vshift;
  int len;
  if(key == 13)  // If Return is pressed
  {
    leading = readicon(handle, 21);  // Read the leading from the icon
    vshift = readicon(handle, 5);  // Read the Vshift from the icon
    textpal_leading(leading);     // Set the leading values for future use
    setleading(leading);         // Set the leading in the text
    len = slen(vshift);         // Extract the vshift value from the string
    textpal_vval = stoi(mids(vshift, 0, len-1));//Set the global vshift value
    setbaseshift(textpal_vval*65536/100);       //Set the vshift in the text
  }
  return(0);
}

// Initialize and add an entry for this applet to the applet menu

void main(void)
{
  string help = "{TEXTPAL_01}";
  translate(help);
  script_menu_initialise();  // Ensure that the Applets menu exists
  script_menu_insert("textpal_entry","textpal_tick","","","{TEXTPAL_00}");  // Then add the 'Text palette' entry to it


  //Add event handlers
  addeventhandler(0x212, 0, "textpal_caret");       //Entering/leaving a view
  addeventhandler(0x304, 0, "textpal_textupdate");  //Text updated
  addeventhandler(0x215, 0, "textpal_activeview");
  //Create a button macro for the applet that will open the palette
  if(isdefmacro("textpal") == -1) defmacro3(12,"textpal","{textpal_entry(0,0)}","textpal",help,"textpal");
}
